from uuid import UUID

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.scoping import ScopedSession
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.sql.schema import Column, Sequence
from sqlalchemy.sql.sqltypes import Integer, String, BigInteger

from eventsourcing.infrastructure.stored_events.base import StoredEventRepository
from eventsourcing.infrastructure.stored_events.transcoders import StoredEvent
from eventsourcing.utils.time import timestamp_long_from_uuid


def get_scoped_session_facade(uri):
    if uri is None: raise ValueError
    engine = create_engine(uri)
    Base.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)
    scoped_session_facade = scoped_session(session_factory)
    return scoped_session_facade


Base = declarative_base()


class SqlStoredEvent(Base):

    __tablename__ = 'stored_events'

    id = Column(Integer, Sequence('stored_event_id_seq'), primary_key=True)
    event_id = Column(String(), index=True)
    timestamp_long = Column(BigInteger(), index=True)
    stored_entity_id = Column(String(), index=True)
    event_topic = Column(String())
    event_attrs = Column(String())

    # def __repr__(self):
    #     return "<SqlStoredEvent(id='%s', event_id='%s', entity_id='%s', event_topic='%s', event_attrs='%s')>" % (
    #         self.id, self.event_id, self.entity_id, self.event_topic, self.event_attrs)


def from_sql(sql_stored_event):
    assert isinstance(sql_stored_event, SqlStoredEvent), sql_stored_event
    return StoredEvent(
        event_id=sql_stored_event.event_id,
        stored_entity_id=sql_stored_event.stored_entity_id,
        event_attrs=sql_stored_event.event_attrs,
        event_topic=sql_stored_event.event_topic
    )


def to_sql(stored_event):
    assert isinstance(stored_event, StoredEvent)
    return SqlStoredEvent(
        event_id=stored_event.event_id,
        timestamp_long=timestamp_long_from_uuid(stored_event.event_id),
        stored_entity_id=stored_event.stored_entity_id,
        event_attrs=stored_event.event_attrs,
        event_topic=stored_event.event_topic
    )


class SQLAlchemyStoredEventRepository(StoredEventRepository):

    serialize_with_uuid1 = True

    def __init__(self, db_session, **kwargs):
        super(SQLAlchemyStoredEventRepository, self).__init__(**kwargs)
        assert isinstance(db_session, ScopedSession)
        self.db_session = db_session

    def append(self, stored_event):
        sql_stored_event = to_sql(stored_event)
        try:
            self.db_session.add(sql_stored_event)
            self.db_session.commit()
        except:
            self.db_session.rollback()
            raise
        finally:
            self.db_session.close() # Begins a new transaction

    # def __contains__(self, pk):
    #     (stored_entity_id, event_id) = pk
    #     query = self.db_session.query(SqlStoredEvent)
    #     query = query.filter_by(stored_entity_id=stored_entity_id, event_id=event_id.hex)
    #     return bool(query.count())
    #
    # def __getitem__(self, pk):
    #     (stored_entity_id, event_id) = pk
    #     query = self.db_session.query(SqlStoredEvent)
    #     query = query.filter_by(stored_entity_id=stored_entity_id, event_id=event_id.hex)
    #     sql_stored_event = query.first()
    #     if sql_stored_event is not None:
    #         return from_sql(sql_stored_event)
    #     else:
    #         raise KeyError
    #
    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_asc=False):
        query = self.db_session.query(SqlStoredEvent)
        query = query.filter_by(stored_entity_id=stored_entity_id)
        if query_asc:
            query = query.order_by(asc(SqlStoredEvent.id))
        else:
            query = query.order_by(desc(SqlStoredEvent.id))
        if until is not None:
            query = query.filter(SqlStoredEvent.timestamp_long <= timestamp_long_from_uuid(until))
        if after is not None:
            query = query.filter(SqlStoredEvent.timestamp_long > timestamp_long_from_uuid(after))
        if limit is not None:
            query = query.limit(limit)
        events = self.map(from_sql, query)
        if not query_asc:
            events = reversed(list(events))
        return events

    def get_most_recent_event(self, stored_entity_id, until=None):
        query = self.db_session.query(SqlStoredEvent)
        query = query.filter_by(stored_entity_id=stored_entity_id)
        query = query.order_by(desc(SqlStoredEvent.id))
        if until is not None:
            query = query.filter(SqlStoredEvent.timestamp_long <= timestamp_long_from_uuid(until))
        sql_stored_event = query.first()
        return from_sql(sql_stored_event) if sql_stored_event else None



import unittest
from eventsourcing.application.example.base import ExampleApplication
from eventsourcing.domain.model.example import Example
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepo
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber
from eventsourcing.infrastructure.stored_events.base import StoredEventRepository

__author__ = 'john'


class ExampleApplicationTestCase(unittest.TestCase):

    def assert_is_example_application(self, app):
        assert isinstance(app, ExampleApplication)  # For PyCharm...

        # Check there's a stored event repo.
        self.assertIsInstance(app.stored_event_repo, StoredEventRepository)

        # Check there's an event store.
        self.assertIsInstance(app.event_store, EventStore)
        self.assertEqual(app.event_store.stored_event_repo, app.stored_event_repo)

        # Check there's a persistence subscriber.
        self.assertIsInstance(app.persistence_subscriber, PersistenceSubscriber)
        self.assertEqual(app.persistence_subscriber.event_store, app.event_store)

        # Check there's an example repository.
        self.assertIsInstance(app.example_repo, ExampleRepo)

        # Register a new example.
        example1 = app.register_new_example(a=10, b=20)
        self.assertIsInstance(example1, Example)

        # Check the example is available in the repo.
        entity1 = app.example_repo[example1.id]
        self.assertEqual(10, entity1.a)
        self.assertEqual(20, entity1.b)
        self.assertEqual(example1, entity1)

        # Change attribute values.
        entity1.a = 100

        # Check the new value is available in the repo.
        entity1 = app.example_repo[example1.id]
        self.assertEqual(100, entity1.a)


from eventsourcing.application.example.with_sqlalchemy import ExampleApplicationWithSQLAlchemy
from eventsourcingtests.example_application_testcase import ExampleApplicationTestCase


class TestApplicationWithSQLAlchemy(ExampleApplicationTestCase):
    def test_application_with_sqlalchemy(self):
        # Setup the example application, use it as a context manager.
        with ExampleApplicationWithSQLAlchemy(db_uri='sqlite:///:memory:') as app:
            self.assert_is_example_application(app)



import unittest
try:
    from unittest import mock
except:
    import mock
from eventsourcing.domain.model.events import publish, DomainEvent
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber


class TestPersistenceSubscriber(unittest.TestCase):

    def setUp(self):
        # Set up a persistence subscriber with a (mock) event store.
        self.event_store = mock.Mock(spec=EventStore)
        self.persistence_subscriber = PersistenceSubscriber(event_store=self.event_store)

    def tearDown(self):
        # Close the persistence subscriber.
        self.persistence_subscriber.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        self.assertEqual(0, self.event_store.append.call_count)

        # Publish a (mock) domain event.
        domain_event = mock.Mock(spec=DomainEvent)
        publish(domain_event)

        # Check the append method HAS been called once with the domain event.
        self.event_store.append.assert_called_once_with(domain_event)



from abc import abstractmethod, ABCMeta

from six import with_metaclass

from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.persistence_subscriber import PersistenceSubscriber


class EventSourcingApplication(with_metaclass(ABCMeta)):

    def __init__(self, json_encoder_cls=None, json_decoder_cls=None):
        self.stored_event_repo = self.create_stored_event_repo(json_encoder_cls=json_encoder_cls,
                                                               json_decoder_cls=json_decoder_cls)
        self.event_store = self.create_event_store()
        self.persistence_subscriber = self.create_persistence_subscriber()

    @abstractmethod
    def create_stored_event_repo(self, **kwargs):
        """Returns an instance of a subclass of StoredEventRepository.

        :rtype: StoredEventRepository
        """

    def create_event_store(self):
        return EventStore(self.stored_event_repo)

    def create_persistence_subscriber(self):
        return PersistenceSubscriber(self.event_store)

    def close(self):
        self.persistence_subscriber.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()



import os

from cassandra import ConsistencyLevel, AlreadyExists
from cassandra.auth import PlainTextAuthProvider
from cassandra.cqlengine.models import Model, columns
from cassandra.cqlengine.management import sync_table, create_keyspace_simple
import cassandra.cqlengine.connection
import six

from eventsourcing.infrastructure.stored_events.base import StoredEventRepository, ThreadedStoredEventIterator
from eventsourcing.infrastructure.stored_events.transcoders import StoredEvent


class CqlStoredEvent(Model):
    __table_name__ = 'stored_events'
    n = columns.Text(partition_key=True)  # 'n' is for the stored entity ID
    v = columns.TimeUUID(clustering_order='DESC', primary_key=True)  # 'v' is for event ID
    t = columns.Text(required=True)  # 't' is for event topic
    a = columns.Text(required=True)  # 'a' is for event attributes


def to_cql(stored_event):
    assert isinstance(stored_event, StoredEvent)
    return CqlStoredEvent(
        n=stored_event.stored_entity_id,
        v=stored_event.event_id,
        t=stored_event.event_topic,
        a=stored_event.event_attrs
    )


def from_cql(cql_stored_event):
    assert isinstance(cql_stored_event, CqlStoredEvent), cql_stored_event
    return StoredEvent(
        stored_entity_id=cql_stored_event.n,
        event_id=cql_stored_event.v.hex,
        event_topic=cql_stored_event.t,
        event_attrs=cql_stored_event.a
    )


class CassandraStoredEventRepository(StoredEventRepository):

    serialize_with_uuid1 = True

    @property
    def iterator_class(self):
        return ThreadedStoredEventIterator

    def append(self, stored_event):
        cql_stored_event = to_cql(stored_event)
        cql_stored_event.save()

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_asc=False):
        query = CqlStoredEvent.objects.filter(n=stored_entity_id)
        if query_asc:
            query = query.order_by('v')
        if until is not None:
            query = query.filter(v__lte=until)
        if after is not None:
            query = query.filter(v__gt=after)
        if limit is not None:
            query = query.limit(limit)
        events = self.map(from_cql, query)
        if not query_asc:
            events = reversed(list(events))
        return events


def get_cassandra_setup_params(hosts=('localhost',), consistency='QUORUM', default_keyspace='eventsourcing', port=9042,
                               protocol_version=3, username=None, password=None):

    # Construct an "auth provider" object.
    if username and password:
        auth_provider = PlainTextAuthProvider(username, password)
    else:
        auth_provider = None

    # Resolve the consistency level to an object, if it's a string.
    if isinstance(consistency, six.string_types):
        try:
            consistency = getattr(ConsistencyLevel, consistency.upper())
        except AttributeError:
            msg = "Cassandra consistency level '{}' not found.".format(consistency)
            raise Exception(msg)

    return auth_provider, hosts, consistency, default_keyspace, port, protocol_version


def setup_cassandra_connection(auth_provider, hosts, consistency, default_keyspace, port, protocol_version):
    cassandra.cqlengine.connection.setup(
        hosts=hosts,
        default_keyspace=default_keyspace,
        consistency=consistency,
        # port=port,
        # auth_provider=auth_provider,
        # protocol_version=protocol_version,
        # lazy_connect=True,
    )


def create_cassandra_keyspace_and_tables(default_keyspace, replication_factor=1):
    os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'
    try:
        create_keyspace_simple(default_keyspace, replication_factor=replication_factor)
    except AlreadyExists:
        pass
    else:
        sync_table(CqlStoredEvent)


def shutdown_cassandra_connection():
    if cassandra.cqlengine.connection.session:
        cassandra.cqlengine.connection.session.shutdown()
    if cassandra.cqlengine.connection.cluster:
        cassandra.cqlengine.connection.cluster.shutdown()



from abc import abstractproperty

from eventsourcing.domain.model.entity import EntityRepository, EventSourcedEntity
from eventsourcing.infrastructure.event_player import EventPlayer, entity_from_snapshot
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events.transcoders import id_prefix_from_entity_class


class EventSourcedRepository(EntityRepository):

    def __init__(self, event_store, use_cache=False):
        assert isinstance(event_store, EventStore)
        self.event_store = event_store
        domain_class = self.domain_class
        assert issubclass(domain_class, EventSourcedEntity)
        self.event_player = EventPlayer(
            event_store=event_store,
            id_prefix=id_prefix_from_entity_class(domain_class),
            mutate_func=domain_class.mutate,
            page_size=domain_class.__page_size__
        )
        self._cache = {}
        self._use_cache = use_cache

    @abstractproperty
    def domain_class(self):
        """
        Returns the type of entity held by this repository.
        """

    def __contains__(self, entity_id):
        """
        Returns a boolean value according to whether or the entity with given ID exists.
        """
        try:
            self.__getitem__(entity_id)
        except KeyError:
            return False
        else:
            return True

    def __getitem__(self, entity_id):
        """
        Returns entity with given ID.
        """
        # Get entity from the cache.
        if self._use_cache:
            try:
                return self._cache[entity_id]
            except KeyError:
                pass

        # Reconstitute the entity.
        entity = self.get_entity(entity_id)

        # Never created or already discarded?
        if entity is None:
            raise KeyError(entity_id)

        # Put entity in the cache.
        if self._use_cache:
             self._cache[entity_id] = entity

        # Return entity.
        return entity

    def get_entity(self, entity_id, until=None):
        """
        Returns entity with given ID, optionally as it was at the given time.
        """
        # Get a snapshot (None if none exist).
        snapshot = self.event_player.get_snapshot(entity_id, until)

        # Decide the initial state, and after when we need to get the events.
        if snapshot is None:
            after = None
            initial_state = None
        else:
            after = snapshot.domain_event_id
            initial_state = entity_from_snapshot(snapshot)

        # Replay domain events.
        return self.event_player.replay_events(entity_id, after=after, until=until, initial_state=initial_state)



from functools import reduce

from eventsourcing.domain.model.snapshot import take_snapshot
from eventsourcing.domain.services.snapshot import get_snapshot
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events.transcoders import deserialize_domain_entity, make_stored_entity_id


class EventPlayer(object):

    def __init__(self, event_store, id_prefix, mutate_func, page_size=None):
        assert isinstance(event_store, EventStore), event_store
        self.event_store = event_store
        self.id_prefix = id_prefix
        self.mutate = mutate_func
        self.page_size = page_size

    def replay_events(self, entity_id, after=None, until=None, initial_state=None):
        # Make the stored entity ID.
        stored_entity_id = self.make_stored_entity_id(entity_id)

        # Get entity's domain events from the event store.
        domain_events = self.event_store.get_entity_events(
            stored_entity_id=stored_entity_id,
            after=after,
            until=until,
            page_size=self.page_size
        )

        # Duplicate initial state, to preserve state of given object.
        if initial_state is not None:
            initial_state_copy = object.__new__(type(initial_state))
            initial_state_copy.__dict__.update(initial_state.__dict__)
            initial_state = initial_state_copy

        # Mutate initial state using the sequence of domain events.
        domain_entity = reduce(self.mutate, domain_events, initial_state)

        # Return the resulting domain entity.
        return domain_entity

    def take_snapshot(self, entity_id, until=None):
        """
        Takes a snapshot of the entity as it existed after the most recent event, optionally until a given time.
        """
        # Get the ultimate event (until a particular time).
        most_recent_event = self.get_most_recent_event(entity_id, until=until)

        # If there isn't any event, there's nothing to snapshot, so return None.
        if most_recent_event is None:
            return None

        until_event_id = most_recent_event.domain_event_id

        # If there is something to snapshot, then get the ultimate snapshot (until a particular time).
        last_snapshot = self.get_snapshot(entity_id, until=until_event_id)

        # If there's a snapshot...
        if last_snapshot:
            if last_snapshot.at_event_id == until_event_id:
                # and it is coincident with the last event, then there's nothing to do.
                return last_snapshot
            else:
                # otherwise there must be events after the snapshot.
                after_event_id = last_snapshot.at_event_id
                initial_state = entity_from_snapshot(last_snapshot)
        else:
            # We need to get all the events (until a particular time) and start from scratch.
            after_event_id = None
            initial_state = None

        # Get entity in the state after this event was applied.
        entity = self.replay_events(entity_id, after=after_event_id, until=until_event_id, initial_state=initial_state)

        # Take a snapshot of the entity exactly when this event was applied.
        return take_snapshot(entity, at_event_id=until_event_id)

    def get_snapshot(self, entity_id, until=None):
        """
        Returns a snapshot of the entity, optional as it existed until a given time.
        """
        stored_entity_id = self.make_stored_entity_id(entity_id=entity_id)
        return get_snapshot(stored_entity_id, self.event_store, until=until)

    def get_most_recent_event(self, entity_id, until=None):
        """
        Returns the most recent event for the given entity ID.
        """
        stored_entity_id = self.make_stored_entity_id(entity_id)
        return self.event_store.get_most_recent_event(stored_entity_id, until=until)

    def make_stored_entity_id(self, entity_id):
        "Prefixes the given entity ID with the ID prefix for entity's type."
        return make_stored_entity_id(self.id_prefix, entity_id)


def entity_from_snapshot(snapshot):
    return deserialize_domain_entity(snapshot.topic, snapshot.attrs)



from eventsourcing.infrastructure.stored_events.python_objects_stored_events import PythonObjectsStoredEventRepository
from eventsourcingtests.test_stored_events import StoredEventRepositoryTestCase


class TestPythonObjectsStoredEventRepository(StoredEventRepositoryTestCase):

    def test_stored_events_in_memory(self):
        self.checkStoredEventRepository(PythonObjectsStoredEventRepository())


from eventsourcing.infrastructure.stored_events.shared_memory_stored_events import SharedMemoryStoredEventRepository
from eventsourcingtests.test_stored_events import StoredEventRepositoryTestCase


class TestSharedMemoryStoredEventRepository(StoredEventRepositoryTestCase):

    def test_stored_events_in_shared_memory(self):
        self.checkStoredEventRepository(SharedMemoryStoredEventRepository())


from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.stored_events.cassandra_stored_events import setup_cassandra_connection, \
    get_cassandra_setup_params, shutdown_cassandra_connection, CassandraStoredEventRepository


DEFAULT_CASSANDRA_KEYSPACE = 'eventsourcing'


class EventSourcingWithCassandra(EventSourcingApplication):

    def __init__(self, hosts=('localhost',), consistency='QUORUM', default_keyspace=DEFAULT_CASSANDRA_KEYSPACE,
                 port=9042, protocol_version=2, username=None, password=None, **kwargs):
        self.setup_cassandra_connection(hosts, consistency, default_keyspace, port, protocol_version, username,
                                        password)
        super(EventSourcingWithCassandra, self).__init__(**kwargs)

    @staticmethod
    def setup_cassandra_connection(*args):
        setup_cassandra_connection(*get_cassandra_setup_params(*args))

    def create_stored_event_repo(self, **kwargs):
        return CassandraStoredEventRepository(**kwargs)

    def close(self):
        super(EventSourcingWithCassandra, self).close()
        shutdown_cassandra_connection()



from unittest.case import TestCase
from uuid import uuid1

import six
from cassandra.cqlengine.management import drop_keyspace
from math import floor

from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects
from eventsourcing.application.example.with_sqlalchemy import ExampleApplicationWithSQLAlchemy
from eventsourcing.application.with_cassandra import DEFAULT_CASSANDRA_KEYSPACE
from eventsourcing.domain.model.example import register_new_example
from eventsourcing.infrastructure.stored_events.cassandra_stored_events import create_cassandra_keyspace_and_tables
from eventsourcing.infrastructure.stored_events.transcoders import make_stored_entity_id
from eventsourcing.utils.time import utc_now


class PerformanceTestCase(TestCase):

    def setUp(self):
        self.skipTest('Abstract test ignored\n')
        self.app = None

    def test_example_performance(self):

        # Initialise dict of entities.
        self.entities = {}

        report_name = type(self).__name__[4:]
        print("\n{} report:\n".format(report_name))

        repetitions = 10

        # NB: Use range(1, 5) to test whether we can get more than 10000 event from Cassandra.
        for i in six.moves.range(1, 5):
            # Setup a number of entities, with different lengths of event history.
            payload = 3
            # payload = str([uuid4().hex for _ in six.moves.range(100000)])
            example = register_new_example(a=1, b=payload)
            self.entities[i] = example

            # Beat a number of times.
            start_beating = utc_now()
            num_beats = int(floor(10 ** i))
            for _ in six.moves.range(num_beats):
                example.beat_heart()
            time_beating = utc_now() - start_beating
            print("Time to beat {} times: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                  "".format(num_beats, time_beating, num_beats / time_beating, time_beating / num_beats))

            # Get the last n events from the repo.
            def last_n(n):
                n = min(n, num_beats + 1)
                stored_entity_id = make_stored_entity_id('Example', example.id)
                repo = self.app.example_repo.event_player.event_store.stored_event_repo

                start_last_n = utc_now()
                last_n_stored_events = []
                for _ in six.moves.range(repetitions):
                    last_n_stored_events = repo.get_most_recent_events(stored_entity_id, limit=n)
                time_last_n = (utc_now() - start_last_n) / repetitions

                num_retrieved_events = len(list(last_n_stored_events))
                events_per_second = num_retrieved_events / time_last_n
                print(("Time to get last {:>"+str(i+1)+"} events after {} events: {:.6f}s ({:.0f} events/s)"
                      "").format(n, num_beats + 1, time_last_n, events_per_second))

            for j in range(0, i+1):
                last_n(10**j)

            # Get the entity by replaying all events (which it must since there isn't a snapshot).
            start_replay = utc_now()
            for _ in six.moves.range(repetitions):
                _ = self.app.example_repo[example.id]
            time_replaying = (utc_now() - start_replay) / repetitions
            print("Time to replay {} beats: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                  "".format(num_beats, time_replaying, num_beats / time_replaying, time_replaying / num_beats))

            # Take snapshot, and beat heart a few more times.
            self.app.example_repo.event_player.take_snapshot(example.id, until=uuid1().hex)

            extra_beats = 4
            for _ in six.moves.range(extra_beats):
                example.beat_heart()
            num_beats += extra_beats

            # Get the entity using snapshot and replaying events since the snapshot.
            start_replay = utc_now()
            for _ in six.moves.range(repetitions):
                _ = self.app.example_repo[example.id]
            time_replaying = (utc_now() - start_replay) / repetitions

            events_per_second = (extra_beats + 1) / time_replaying  # +1 for the snapshot event
            beats_per_second = num_beats / time_replaying
            print("Time to replay snapshot with {} extra beats: {:.6f}s ({:.0f} events/s, {:.0f} beats/s)"
                  "".format(extra_beats, time_replaying, events_per_second, beats_per_second))
            
            print("")


class TestCassandraPerformance(PerformanceTestCase):

    def setUp(self):
        # Setup the example application.
        self.app = ExampleApplicationWithCassandra()

        # Setup the keyspace and column family for stored events.
        create_cassandra_keyspace_and_tables(DEFAULT_CASSANDRA_KEYSPACE)

    def tearDown(self):
        # Drop the keyspace.
        drop_keyspace(DEFAULT_CASSANDRA_KEYSPACE)

        # Close the application.
        self.app.close()


class TestSQLAlchemyPerformance(PerformanceTestCase):

    def setUp(self):
        # Setup the example application.
        self.app = ExampleApplicationWithSQLAlchemy(db_uri='sqlite:///:memory:')

    def tearDown(self):
        # Close the application.
        self.app.close()


class TestPythonObjectsPerformance(PerformanceTestCase):

    def setUp(self):
        # Setup the example application.
        self.app = ExampleApplicationWithPythonObjects()

    def tearDown(self):
        # Close the application.
        self.app.close()



from eventsourcing.infrastructure.stored_events.base import StoredEventRepository
from eventsourcing.infrastructure.stored_events.transcoders import StoredEvent
from eventsourcing.utils.time import timestamp_from_uuid


class PythonObjectsStoredEventRepository(StoredEventRepository):

    serialize_without_json = True
    serialize_with_uuid1 = True

    def __init__(self):
        super(PythonObjectsStoredEventRepository, self).__init__()
        self._by_id = {}
        self._by_stored_entity_id = {}
        self._by_topic = {}

    def append(self, stored_event):
        assert isinstance(stored_event, StoredEvent)
        event_id = stored_event.event_id
        stored_entity_id = stored_event.stored_entity_id
        topic = stored_event.event_topic

        if topic.endswith('Discarded'):

            self.remove_entity(stored_entity_id)

        else:
            # Index by event ID.
            self._by_id[event_id] = stored_event

            # Index by entity ID.
            if stored_entity_id not in self._by_stored_entity_id:
                self._by_stored_entity_id[stored_entity_id] = []
            self._by_stored_entity_id[stored_entity_id].append(stored_event)

            # Index by event topic.
            if topic not in self._by_topic:
                self._by_topic[topic] = []
            self._by_topic[topic].append(stored_event)

    def remove_entity(self, stored_entity_id):
        if stored_entity_id in self._by_stored_entity_id:
            for stored_event in self._by_stored_entity_id.pop(stored_entity_id):
                del(self._by_id[stored_event.event_id])
                if stored_event.event_topic in self._by_topic:
                    self._by_topic[stored_event.event_topic].remove(stored_event)

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_asc=False):
        if stored_entity_id not in self._by_stored_entity_id:
            return []
        else:
            events = []
            count = 0
            stored_events = self._by_stored_entity_id[stored_entity_id]
            if after is None:
                after_timestamp = None
            else:
                after_timestamp = timestamp_from_uuid(after)
            if until is None:
                until_timestamp = None
            else:
                until_timestamp = timestamp_from_uuid(until)
            for event in stored_events:
                event_timestamp = timestamp_from_uuid(event.event_id)
                if after_timestamp:
                    # Exclude if earlier or equal to the 'after' time.
                    if event_timestamp <= after_timestamp:
                        continue
                if until_timestamp:
                    # Exclude if later than the 'until' time.
                    if event_timestamp > until_timestamp:
                        continue
                count += 1
                events.append(event)

            if limit is not None:
                if not query_asc:
                    events.reverse()
                events = events[:min(limit, len(events))]
                if not query_asc:
                    events.reverse()

            return events



from abc import ABCMeta, abstractmethod
from threading import Thread

import six

from eventsourcing.infrastructure.stored_events.transcoders import serialize_domain_event, deserialize_domain_event


class StoredEventRepository(six.with_metaclass(ABCMeta)):

    serialize_without_json = False
    serialize_with_uuid1 = False

    def __init__(self, json_encoder_cls=None, json_decoder_cls=None):
        self.json_encoder_cls = json_encoder_cls
        self.json_decoder_cls = json_decoder_cls

    @abstractmethod
    def append(self, stored_event):
        """Saves given stored event in this repository.
        :param stored_event:
        """

    @abstractmethod
    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_asc=False):
        """Returns all events for given entity ID in chronological order. Limit is max 10000.
        :param until:
        :param after:
        :param stored_entity_id:
        :rtype: list
        """

    def iterate_entity_events(self, stored_entity_id, after=None, until=None, limit=None, page_size=None):
        """Returns all events for given entity ID in chronological by paging through the stored events.
        :param until:
        :param after:
        :param stored_entity_id:
        :rtype: list
        """
        return self.iterator_class(
            repo=self,
            stored_entity_id=stored_entity_id,
            after=after,
            until=until,
            limit=limit,
            page_size=page_size,
        )

    @property
    def iterator_class(self):
        return SimpleStoredEventIterator

    def get_most_recent_event(self, stored_entity_id, until=None):
        """Returns last event for given entity ID.

        :param stored_entity_id:
        :rtype: DomainEvent, NoneType
        """
        events = self.get_most_recent_events(stored_entity_id, until=until, limit=1)
        events = list(events)
        if len(events) == 1:
            return events[0]
        elif len(events) == 0:
            return None
        else:
            raise Exception("Shouldn't have more than one object: {}".format(events))

    def get_most_recent_events(self, stored_entity_id, until=None, limit=None):
        return self.get_entity_events(stored_entity_id, until=until, limit=limit)

    def serialize(self, domain_event):
        """Returns a stored event from a domain event.
        :type domain_event: object
        :param domain_event:
        """
        return serialize_domain_event(
            domain_event,
            json_encoder_cls=self.json_encoder_cls,
            without_json=self.serialize_without_json,
            with_uuid1=self.serialize_with_uuid1
        )

    def deserialize(self, stored_event):
        """Returns a domain event from a stored event.
        :type stored_event: object
        """
        return deserialize_domain_event(
            stored_event,
            json_decoder_cls=self.json_decoder_cls,
            without_json=self.serialize_without_json
        )

    @staticmethod
    def map(func, iterable):
        return six.moves.map(func, iterable)


class StoredEventIterator(six.with_metaclass(ABCMeta)):

    DEFAULT_PAGE_SIZE = 1000

    def __init__(self, repo, stored_entity_id, page_size=None, after=None, until=None, limit=None):
        assert isinstance(repo, StoredEventRepository)
        assert isinstance(stored_entity_id, six.string_types)
        assert isinstance(page_size, (six.integer_types, type(None)))
        assert isinstance(limit, (six.integer_types, type(None)))
        self.repo = repo
        self.stored_entity_id = stored_entity_id
        self.page_size = page_size or self.DEFAULT_PAGE_SIZE
        self.after = after
        self.until = until
        self.limit = limit
        self.page_counter = 0
        self.all_event_counter = 0

    def inc_page_counter(self):
        """
        Increments the page counter.

        Each query result as a page, even if there are no items in the page. This really counts queries.
         - it is easy to divide the number of events by the page size if the "correct" answer is required
         - there will be a difference in the counts when the number of events can be exactly divided by the page
           size, because there is no way to know in advance that a full page is also the last page.
        """
        self.page_counter += 1

    def inc_all_event_counter(self):
        self.all_event_counter += 1

    @abstractmethod
    def __iter__(self):
        pass


class SimpleStoredEventIterator(StoredEventIterator):

    def __iter__(self):
        # Get pages of events until we hit the last page.
        while True:
            # Get next page of events.
            stored_events = self.repo.get_entity_events(self.stored_entity_id,
                                                        after=self.after,
                                                        until=self.until,
                                                        limit=self.page_size,
                                                        query_asc=True
                                                        )
            # Count the page.
            self.inc_page_counter()

            # Start counting events in this page.
            in_page_event_counter = 0

            # Yield each stored event, so long as we aren't over the limit.
            for stored_event in stored_events:

                # Stop if we're over the limit.
                if self.limit and self.all_event_counter >= self.limit:
                    raise StopIteration

                # Count each event.
                self.inc_all_event_counter()
                in_page_event_counter += 1

                # Yield the event.
                yield stored_event

                # Update loop variables.
                self.after = stored_event.event_id

            # Decide if this is the last page.
            is_last_page = in_page_event_counter != self.page_size

            # If that was the last page, then stop iterating.
            if is_last_page:
                raise StopIteration


class ThreadedStoredEventIterator(StoredEventIterator):

    def __iter__(self):
        # Start a thread to get a page of events.
        thread = self.start_thread()

        # Get pages of stored events, until the page isn't full.
        while True:
            # Wait for the next page of events.
            thread.join()

            # Count the page.
            self.inc_page_counter()

            # Get the stored events from the thread.
            stored_events = thread.stored_events

            # Count the number of stored events in this page.
            num_stored_events = len(stored_events)

            # Decide if this is the last page.
            is_last_page = num_stored_events != self.page_size

            if not is_last_page:
                # Update loop variables.
                self.after = stored_events[-1].event_id

                # Start the next thread.
                thread = self.start_thread()

            # Yield each stored event.
            for stored_event in stored_events:

                # Stop if we're over the limit.
                if self.limit and self.all_event_counter >= self.limit:
                    raise StopIteration

                # Count each event.
                self.inc_all_event_counter()

                # Yield the event.
                yield stored_event

            # If that was the last page, then stop iterating.
            if is_last_page:
                raise StopIteration

    def start_thread(self):
        thread = GetEntityEventsThread(
            repo=self.repo,
            stored_entity_id=self.stored_entity_id,
            after=self.after,
            until=self.until,
            page_size=self.page_size,
            query_asc=True
        )
        thread.start()
        return thread


class GetEntityEventsThread(Thread):

    def __init__(self, repo, stored_entity_id, after=None, until=None, page_size=None, query_asc=False, *args, **kwargs):
        super(GetEntityEventsThread, self).__init__(*args, **kwargs)
        self.repo = repo
        self.stored_entity_id = stored_entity_id
        self.after = after
        self.until = until
        self.page_size = page_size
        self.query_asc = query_asc
        self.stored_events = None

    def run(self):
        self.stored_events = list(self.repo.get_entity_events(
            stored_entity_id=self.stored_entity_id,
            after=self.after,
            until=self.until,
            limit=self.page_size,
            query_asc=self.query_asc
        ))



from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.infrastructure.event_store import EventStore
from eventsourcing.infrastructure.stored_events.transcoders import make_stored_entity_id, id_prefix_from_event_class


def get_snapshot(stored_entity_id, event_store, until=None):
    """
    Get the last snapshot for entity.
    :rtype: Snapshot
    """
    assert isinstance(event_store, EventStore)
    snapshot_entity_id = make_stored_entity_id(id_prefix_from_event_class(Snapshot), stored_entity_id)
    return event_store.get_most_recent_event(snapshot_entity_id, until=until)



import datetime
from uuid import UUID

import six

if hasattr(datetime, 'timezone'):
    utc_timezone = datetime.timezone.utc
else:
    from datetime import tzinfo, timedelta

    class UTC(tzinfo):
        def utcoffset(self, date_time):
            return timedelta(0)

        def dst(self, date_time):
            return timedelta(0)
    utc_timezone = UTC()


def utc_now():
    now_datetime = datetime.datetime.now(utc_timezone)
    try:
        now_timestamp = now_datetime.timestamp()
    except AttributeError:
        now_timestamp = (now_datetime - datetime.datetime(1970, 1, 1, tzinfo=utc_timezone)).total_seconds()
    return now_timestamp


def timestamp_from_uuid(uuid_arg):
    "Return a floating point unix timestamp to 6 decimal places."
    return timestamp_long_from_uuid(uuid_arg) / 1e7


def timestamp_long_from_uuid(uuid_arg):
    if isinstance(uuid_arg, six.string_types):
        uuid_arg = UUID(uuid_arg)
    return (uuid_arg.time - 0x01B21DD213814000)



from unittest.case import TestCase
from uuid import uuid1

from eventsourcing.utils.time import utc_now, timestamp_from_uuid


class TestUtils(TestCase):

    def test_timestamp_from_uuid(self):
        until = utc_now()
        uuid = uuid1()
        after = utc_now()
        uuid_timestamp = timestamp_from_uuid(uuid)
        self.assertLess(until, uuid_timestamp)
        self.assertGreater(after, uuid_timestamp)



import six

from eventsourcing.infrastructure.stored_events.base import StoredEventRepository


class EventStore(object):

    def __init__(self, stored_event_repo):
        assert isinstance(stored_event_repo, StoredEventRepository), stored_event_repo
        self.stored_event_repo = stored_event_repo

    def append(self, domain_event):
        # Serialize the domain event.
        stored_event = self.stored_event_repo.serialize(domain_event)

        # Append the stored event to the stored event repo.
        self.stored_event_repo.append(stored_event)

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, page_size=None):
        # Get the events that have been stored for the entity.
        if not page_size:
            stored_events = self.stored_event_repo.get_entity_events(
                stored_entity_id=stored_entity_id,
                after=after,
                until=until,
                limit=limit
            )
        else:
            stored_events = self.stored_event_repo.iterate_entity_events(
                stored_entity_id=stored_entity_id,
                after=after,
                until=until,
                limit=limit,
                page_size=page_size
            )

        # Deserialize all the stored event objects into domain event objects.
        return six.moves.map(self.stored_event_repo.deserialize, stored_events)

    def get_most_recent_event(self, stored_entity_id, until=None):
        """Returns last event for given entity ID.

        :rtype: DomainEvent, NoneType
        """
        stored_event = self.stored_event_repo.get_most_recent_event(stored_entity_id, until=until)
        return None if stored_event is None else self.stored_event_repo.deserialize(stored_event)



from cassandra.cqlengine.management import drop_keyspace

from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.application.with_cassandra import DEFAULT_CASSANDRA_KEYSPACE
from eventsourcing.infrastructure.stored_events.cassandra_stored_events import create_cassandra_keyspace_and_tables
from eventsourcingtests.example_application_testcase import ExampleApplicationTestCase


class TestApplicationWithCassandra(ExampleApplicationTestCase):

    def test_application_with_cassandra(self):
        # Setup the example application, use it as a context manager.
        with ExampleApplicationWithCassandra() as app:
            create_cassandra_keyspace_and_tables(DEFAULT_CASSANDRA_KEYSPACE)
            try:
                self.assert_is_example_application(app)
            finally:
                drop_keyspace(DEFAULT_CASSANDRA_KEYSPACE)



import datetime
import importlib
import json
import uuid
from collections import namedtuple

import dateutil.parser
from six import BytesIO

from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.events import topic_from_domain_class, resolve_domain_topic, resolve_attr, DomainEvent

try:
    import numpy
except:
    numpy = None


def serialize_domain_event(domain_event, json_encoder_cls=None, without_json=False, with_uuid1=False):
    """
    Serializes a domain event into a stored event.
    """
    if json_encoder_cls is None:
        json_encoder_cls = ObjectJSONEncoder
    # assert isinstance(domain_event, DomainEvent)
    if with_uuid1:
        event_id = domain_event.domain_event_id
    else:
        event_id = uuid.uuid4().hex
    stored_entity_id = make_stored_entity_id(id_prefix_from_event(domain_event), domain_event.entity_id)
    event_topic = topic_from_domain_class(domain_event.__class__)
    if without_json:
        event_attrs = domain_event
    else:
        event_attrs = json.dumps(domain_event.__dict__, separators=(',', ':'), sort_keys=True, cls=json_encoder_cls)
    return StoredEvent(
        event_id=event_id,
        stored_entity_id=stored_entity_id,
        event_topic=event_topic,
        event_attrs=event_attrs,
    )


def deserialize_domain_event(stored_event, json_decoder_cls=None, without_json=False):
    """
    Recreates original domain event from stored event topic and event attrs.
    """
    assert isinstance(stored_event, StoredEvent)
    if json_decoder_cls is None:
        json_decoder_cls = ObjectJSONDecoder
    if without_json:
        domain_event = stored_event.event_attrs
    else:
        event_class = resolve_domain_topic(stored_event.event_topic)
        event_data = json.loads(stored_event.event_attrs, cls=json_decoder_cls)
        try:
            domain_event = object.__new__(event_class)
            domain_event.__dict__.update(event_data)
        except TypeError:
            raise TypeError("Unable to instantiate class '{}' with data '{}'".format(stored_event.event_topic, stored_event.event_attrs))
    return domain_event


class ObjectJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            return super(ObjectJSONEncoder, self).default(obj)
        except TypeError as e:
            if "not JSON serializable" not in str(e):
                raise
            if isinstance(obj, datetime.datetime):
                return { 'ISO8601_datetime': obj.strftime('%Y-%m-%dT%H:%M:%S.%f%z') }
            if isinstance(obj, datetime.date):
                return { 'ISO8601_date': obj.isoformat() }
            if numpy is not None and isinstance(obj, numpy.ndarray) and obj.ndim == 1:
                memfile = BytesIO()
                numpy.save(memfile, obj)
                memfile.seek(0)
                serialized = json.dumps(memfile.read().decode('latin-1'))
                d = {
                    '__ndarray__': serialized,
                }
                return d
            else:
                d = {
                    '__class__': obj.__class__.__qualname__,
                    '__module__': obj.__module__,
                }
                return d


class ObjectJSONDecoder(json.JSONDecoder):

    def __init__(self, **kwargs):
        super(ObjectJSONDecoder, self).__init__(object_hook=ObjectJSONDecoder.from_jsonable, **kwargs)

    @staticmethod
    def from_jsonable(d):
        if '__ndarray__' in d:
            return ObjectJSONDecoder._decode_ndarray(d)
        elif '__class__' in d and '__module__' in d:
            return ObjectJSONDecoder._decode_class(d)
        elif 'ISO8601_datetime' in d:
            return ObjectJSONDecoder._decode_datetime(d)
        elif 'ISO8601_date' in d:
            return ObjectJSONDecoder._decode_date(d)
        return d

    @staticmethod
    def _decode_ndarray(d):
        serialized = d['__ndarray__']
        memfile = BytesIO()
        memfile.write(json.loads(serialized).encode('latin-1'))
        memfile.seek(0)
        return numpy.load(memfile)

        # return numpy.array(obj_data, d['dtype']).reshape(d['shape'])

    @staticmethod
    def _decode_class(d):
        class_name = d.pop('__class__')
        module_name = d.pop('__module__')
        module = importlib.import_module(module_name)
        cls = resolve_attr(module, class_name)
        try:
            obj = cls(**d)
        except Exception:
            obj = cls()
            for attr, value in d.items():
                obj.__dict__[attr] = ObjectJSONDecoder.from_jsonable(value)
        return obj

    @staticmethod
    def _decode_date(d):
        return datetime.datetime.strptime(d['ISO8601_date'], '%Y-%m-%d').date()

    @staticmethod
    def _decode_datetime(d):
        return dateutil.parser.parse(d['ISO8601_datetime'])


StoredEvent = namedtuple('StoredEvent', ['event_id', 'stored_entity_id', 'event_topic', 'event_attrs'])


def deserialize_domain_entity(entity_topic, entity_attrs):
    """
    Return a new domain entity object from a given topic (a string) and attributes (a dict).
    """

    # Get the domain entity class from the entity topic.
    domain_class = resolve_domain_topic(entity_topic)

    # Instantiate the domain entity class.
    entity = object.__new__(domain_class)

    # Set the attributes.
    entity.__dict__.update(entity_attrs)

    # Return a new domain entity object.
    return entity


def make_stored_entity_id(id_prefix, entity_id):
    return id_prefix + '::' + entity_id


def id_prefix_from_event(domain_event):
    assert isinstance(domain_event, DomainEvent), type(domain_event)
    return id_prefix_from_event_class(type(domain_event))


def id_prefix_from_event_class(domain_event_class):
    assert issubclass(domain_event_class, DomainEvent), type(domain_event_class)
    return domain_event_class.__qualname__.split('.')[0]


def id_prefix_from_entity(domain_entity):
    assert isinstance(domain_entity, EventSourcedEntity)
    return id_prefix_from_entity_class(type(domain_entity))


def id_prefix_from_entity_class(domain_class):
    assert issubclass(domain_class, EventSourcedEntity)
    return domain_class.__name__



import datetime
import unittest
import uuid

import six

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.domain.model.example import Example
from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.infrastructure.stored_events.base import SimpleStoredEventIterator
from eventsourcing.infrastructure.stored_events.transcoders import serialize_domain_event, deserialize_domain_event, \
    resolve_domain_topic, StoredEvent, ObjectJSONDecoder, ObjectJSONEncoder
from eventsourcing.utils.time import utc_timezone


class TestStoredEvent(unittest.TestCase):

    def test_serialize_domain_event(self):
        datetime_now = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429)
        datetime_now_tzaware = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429, tzinfo=utc_timezone)
        date_now = datetime.date(2015, 9, 8)
        event1 = DomainEvent(a=1, b=2, c=datetime_now, d=datetime_now_tzaware, e=date_now, entity_version=0, entity_id='entity1', domain_event_id=3)
        stored_event = serialize_domain_event(event1, json_encoder_cls=ObjectJSONEncoder)
        self.assertEqual('DomainEvent::entity1', stored_event.stored_entity_id)
        self.assertEqual('eventsourcing.domain.model.events#DomainEvent', stored_event.event_topic)
        self.assertEqual('{"a":1,"b":2,"c":{"ISO8601_datetime":"2015-09-08T16:20:50.577429"},"d":{"ISO8601_datetime":"2015-09-08T16:20:50.577429+0000"},"domain_event_id":3,"e":{"ISO8601_date":"2015-09-08"},"entity_id":"entity1","entity_version":0}',
                         stored_event.event_attrs)

    def test_serialize_domain_event_with_numpy_array(self):
        try:
            import numpy
        except ImportError:
            numpy = None

        if numpy is not None:
            event1 = DomainEvent(a=numpy.array([10.123456]), entity_version=0, entity_id='entity1', domain_event_id=3)

            stored_event = serialize_domain_event(event1, json_encoder_cls=ObjectJSONEncoder)
            self.assertEqual('eventsourcing.domain.model.events#DomainEvent', stored_event.event_topic)
            self.assertEqual('{"a":{"__ndarray__":"\\"\\\\u0093NUMPY\\\\u0001\\\\u0000F\\\\u0000{\'descr\': \'<f8\', \'fortran_order\': False, \'shape\': (1,), }            \\\\nm\\\\u00fd\\\\u00f4\\\\u009f5?$@\\""},"domain_event_id":3,"entity_id":"entity1","entity_version":0}',
                             stored_event.event_attrs)
        else:
            self.skipTest("Numpy not installed")

    def test_recreate_domain_event(self):
        stored_event = StoredEvent(event_id='1',
                                   stored_entity_id='entity1',
                                   event_topic='eventsourcing.domain.model.events#DomainEvent',
                                   event_attrs='{"a":1,"b":2,"c":{"ISO8601_datetime":"2015-09-08T16:20:50.577429"},"d":{"ISO8601_datetime":"2015-09-08T16:20:50.577429+0000"},"domain_event_id":3,"e":{"ISO8601_date":"2015-09-08"},"entity_id":"entity1","entity_version":0}')
        domain_event = deserialize_domain_event(stored_event, json_decoder_cls=ObjectJSONDecoder)
        self.assertIsInstance(domain_event, DomainEvent)
        self.assertEqual('entity1', domain_event.entity_id)
        self.assertEqual(1, domain_event.a)
        self.assertEqual(2, domain_event.b)
        datetime_now_timezoneaware = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429, tzinfo=utc_timezone)
        # self.assertEqual(datetime_now, domain_event.c)
        self.assertEqual(datetime_now_timezoneaware, domain_event.d)
        date_now = datetime.date(2015, 9, 8)
        self.assertEqual(date_now, domain_event.e)
        self.assertEqual(3, domain_event.domain_event_id)

        # Check the TypeError is raised.
        stored_event = StoredEvent(event_id='1',
                                   stored_entity_id='entity1',
                                   event_topic='os#path',
                                   event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":3}')
        self.assertRaises(TypeError, deserialize_domain_event, stored_event, json_decoder_cls=ObjectJSONDecoder)

    def test_resolve_event_topic(self):
        example_topic = 'eventsourcing.domain.model.example#Example.Created'
        actual = resolve_domain_topic(example_topic)
        self.assertEqual(Example.Created, actual)
        example_topic = 'xxxxxxxxxxxxx#Example.Event'
        self.assertRaises(TopicResolutionError, resolve_domain_topic, example_topic)
        example_topic = 'eventsourcing.domain.model.example#Xxxxxxxx.Xxxxxxxx'
        self.assertRaises(TopicResolutionError, resolve_domain_topic, example_topic)


class StoredEventRepositoryTestCase(unittest.TestCase):

    def checkStoredEventRepository(self, stored_event_repo):
        stored_entity_id = 'Entity::entity1'

        # Check the repo returns None for calls to get_most_recent_event() when there aren't any events.
        self.assertIsNone(stored_event_repo.get_most_recent_event(stored_entity_id))

        # # Check the repo doesn't contain an event.
        # pk = (stored_entity_id, domain_event_id.uuid1())
        # self.assertNotIn(pk, stored_event_repo)  # __contains__
        # self.assertRaises(KeyError, stored_event_repo.__getitem__, pk)  # __getitem__

        # Store an event for 'entity1'.
        stored_event1 = StoredEvent(event_id=uuid.uuid1().hex,
                                    stored_entity_id=stored_entity_id,
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":3}')
        stored_event_repo.append(stored_event1)

        # # Check the repo contains the event.
        # pk = (stored_entity_id, stored_event1.event_id)
        # self.assertIn(pk, stored_event_repo)  # __contains__
        # self.assertIsInstance(stored_event_repo[pk], StoredEvent)  # __getitem__
        # self.assertIsInstance(repr(stored_event_repo[pk]), str)  # __getitem__

        # Store another event for 'entity1'.
        stored_event2 = StoredEvent(event_id=uuid.uuid1().hex,
                                    stored_entity_id=stored_entity_id,
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":4}')
        stored_event_repo.append(stored_event2)

        # Get all events for 'entity1'.
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.

        num_fixture_events = 2

        self.assertEqual(num_fixture_events, len(retrieved_events))
        # - check the first event
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)
        # - check the second event
        self.assertIsInstance(retrieved_events[1], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[1].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[1].event_attrs)

        # Get with limit (depends on query order).
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, query_asc=True, limit=1)
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, query_asc=True, limit=1)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        # - check the first retrieved event is the first event that was stored
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)

        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, query_asc=False, limit=1)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        # - check the first retrieved event is the last event that was stored
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[0].event_attrs)

        # Get the most recent event for 'entity1'.
        most_recent_event = stored_event_repo.get_most_recent_event(stored_entity_id)
        self.assertIsInstance(most_recent_event, StoredEvent)
        self.assertEqual(most_recent_event.event_id, stored_event2.event_id)

        # Get all events for 'entity1' since the first event's timestamp.
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, after=stored_event1.event_id)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        self.assertEqual(1, len(list(retrieved_events)))
        # - check the last event is first
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[0].event_attrs)

        # Store lots of events, and check the latest events can be retrieved quickly.
        stored_events = []
        num_extra_events = 127
        for page_count in six.moves.range(num_extra_events):
            stored_event_i = StoredEvent(event_id=uuid.uuid1().hex,
                                         stored_entity_id=stored_entity_id,
                                         event_topic='eventsourcing.domain.model.example#Example.Created',
                                         event_attrs='{"a":1,"b":2,"entity_id":"entity1","timestamp":%s}' % (page_count+10))
            stored_events.append(stored_event_i)
            stored_event_repo.append(stored_event_i)

        last_snapshot_event = stored_events[-20]

        # start_time = datetime.datetime.now()
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, after=last_snapshot_event.event_id)
        retrieved_events = list(retrieved_events)
        # page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # self.assertLess(page_duration, 0.05)
        # print("Duration: {}".format(page_duration))

        self.assertEqual(len(retrieved_events), 19)
        self.assertIsInstance(retrieved_events[-1], StoredEvent)
        self.assertEqual(stored_events[-19].event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_events[-19].event_attrs, retrieved_events[0].event_attrs)
        self.assertEqual(stored_events[-1].event_topic, retrieved_events[-1].event_topic)
        self.assertEqual(stored_events[-1].event_attrs, retrieved_events[-1].event_attrs)

        # Check the first events can be retrieved easily.
        start_time = datetime.datetime.now()
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id, query_asc=True, limit=20)
        retrieved_events = list(retrieved_events)
        page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # print("Duration: {}".format(page_duration))
        # self.assertLess(page_duration, 0.05)
        self.assertEqual(len(retrieved_events), 20)
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)

        # Check the next page of events can be retrieved easily.
        start_time = datetime.datetime.now()
        retrieved_events = stored_event_repo.get_entity_events(stored_entity_id,
                                                               after=retrieved_events[-1].event_id,
                                                               limit=20,
                                                               query_asc=True)
        retrieved_events = list(retrieved_events)
        page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # self.assertLess(page_duration, 0.05)
        # print("Duration: {}".format(page_duration))

        self.assertEqual(len(retrieved_events), 20)
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_events[18].event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_events[18].event_attrs, retrieved_events[0].event_attrs)

        # Check the stored event iterator can get all the events.
        # Todo: Move this test of the SimpleStoredEventIterator to a separate test case?
        # Todo: Write a test case for ThreadedStoredEventIterator.
        # start_time = datetime.datetime.now()
        page_size = 50
        iterator = SimpleStoredEventIterator(stored_event_repo, stored_entity_id, page_size=page_size)
        retrieved_events = list(iterator)
        self.assertGreater(len(retrieved_events), page_size)

        self.assertEqual(iterator.page_counter, 3)
        self.assertEqual(iterator.all_event_counter, num_fixture_events + num_extra_events)

        # Check the first and last stored event.
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)
        self.assertEqual(stored_events[-1].event_attrs, retrieved_events[-1].event_attrs)

        # duration = (datetime.datetime.now() - start_time).total_seconds()

        # print("Total duration: {}".format(duration))
        # average_item_duration = duration / len(retrieved_events)

        # print("Average item duration: {}".format(average_item_duration))
        # print("Average item rate: {}".format(1.0 / average_item_duration))
        # self.assertLess(average_item_duration, 0.0005)



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.domain.model.example import register_new_example
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepo


class ExampleApplication(EventSourcingApplication):
    """
    Abstract example event sourced application.

    This application has an Example repository, and a factory method to construct new Example entities.

    It doesn't have a stored event repository.
    """
    def __init__(self, **kwargs):
        super(ExampleApplication, self).__init__(**kwargs)
        self.example_repo = ExampleRepo(self.event_store)

    def register_new_example(self, a, b):
        return register_new_example(a=a, b=b)



from eventsourcing.domain.model.events import DomainEvent, publish
from eventsourcing.domain.model.example import Example
from eventsourcing.infrastructure.stored_events.transcoders import topic_from_domain_class, make_stored_entity_id, \
    id_prefix_from_entity


class Snapshot(DomainEvent):

    def __init__(self, entity_id, topic, attrs, domain_event_id):
        super(Snapshot, self).__init__(entity_id=entity_id,
                                       topic=topic,
                                       attrs=attrs,
                                       domain_event_id=domain_event_id,
                                       entity_version=0)

    @property
    def topic(self):
        """Path to the class.
        """
        return self.__dict__['topic']

    @property
    def attrs(self):
        """Attributes of the instance.
        """
        return self.__dict__['attrs']

    @property
    def at_event_id(self):
        return self.__dict__['domain_event_id']


def take_snapshot(entity, at_event_id):
    # Make the 'stored entity ID' for the entity.
    id_prefix = id_prefix_from_entity(entity)
    stored_snapshotted_entity_id = make_stored_entity_id(id_prefix, entity.id)

    # Create the snapshot event.
    snapshot = Snapshot(
        entity_id=stored_snapshotted_entity_id,
        domain_event_id=at_event_id,
        topic=topic_from_domain_class(Example),
        attrs=entity.__dict__.copy(),
    )
    publish(snapshot)

    # Return the event.
    return snapshot



from eventsourcing.infrastructure.stored_events.sqlalchemy_stored_events import SQLAlchemyStoredEventRepository, \
    get_scoped_session_facade
from eventsourcingtests.test_stored_events import StoredEventRepositoryTestCase


class TestSQLAlchemyStoredEventRepository(StoredEventRepositoryTestCase):

    def test_stored_events_in_sqlalchemy(self):
        stored_event_repo = SQLAlchemyStoredEventRepository(get_scoped_session_facade('sqlite:///:memory:'))
        self.checkStoredEventRepository(stored_event_repo)



from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects
from eventsourcingtests.example_application_testcase import ExampleApplicationTestCase


class TestApplicationWithPythonObjects(ExampleApplicationTestCase):

    def test_application_with_python_objects(self):
        # Setup the example application, use it as a context manager.
        with ExampleApplicationWithPythonObjects() as app:
            self.assert_is_example_application(app)



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.stored_events.sqlalchemy_stored_events import \
    get_scoped_session_facade, SQLAlchemyStoredEventRepository


class EventSourcingWithSQLAlchemy(EventSourcingApplication):

    def __init__(self, db_session=None, db_uri=None, **kwargs):
        self.db_uri = db_uri
        self.db_session = db_session if (db_session is not None) else self.create_db_session(db_uri)
        super(EventSourcingWithSQLAlchemy, self).__init__(**kwargs)

    @staticmethod
    def create_db_session(uri):
        return get_scoped_session_facade(uri)

    def create_stored_event_repo(self, **kwargs):
        return SQLAlchemyStoredEventRepository(db_session=self.db_session, **kwargs)

    def close(self):
        super(EventSourcingWithSQLAlchemy, self).close()
        self.db_session.close()



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.stored_events.python_objects_stored_events import PythonObjectsStoredEventRepository


class EventSourcingWithPythonObjects(EventSourcingApplication):

    def create_stored_event_repo(self, **kwargs):
        return PythonObjectsStoredEventRepository()



from cassandra.cqlengine.management import drop_keyspace

from eventsourcing.application.with_cassandra import DEFAULT_CASSANDRA_KEYSPACE
from eventsourcing.infrastructure.stored_events.cassandra_stored_events import CassandraStoredEventRepository, \
    setup_cassandra_connection, shutdown_cassandra_connection, get_cassandra_setup_params, \
    create_cassandra_keyspace_and_tables
from eventsourcingtests.test_stored_events import StoredEventRepositoryTestCase


class TestCassandraStoredEventRepository(StoredEventRepositoryTestCase):

    def setUp(self):
        super(TestCassandraStoredEventRepository, self).setUp()
        setup_cassandra_connection(*get_cassandra_setup_params(default_keyspace=DEFAULT_CASSANDRA_KEYSPACE))
        create_cassandra_keyspace_and_tables(DEFAULT_CASSANDRA_KEYSPACE)

    def tearDown(self):
        drop_keyspace(DEFAULT_CASSANDRA_KEYSPACE)
        shutdown_cassandra_connection()
        super(TestCassandraStoredEventRepository, self).tearDown()

    def test_stored_events_in_cassandra(self):
        stored_event_repo = CassandraStoredEventRepository()
        self.checkStoredEventRepository(stored_event_repo)
